/* SPDX-License-Identifier: BSD-3-Clause
 * Copyright(c) 2015-2020
 */


#include "base/txgbe.h"
#include "txgbe_ethdev.h"
#include "rte_pmd_txgbe.h"

int
rte_pmd_txgbe_set_vf_mac_addr(uint16_t port, uint16_t vf,
			      struct ether_addr *mac_addr)
{
	struct txgbe_hw *hw = NULL;
	struct txgbe_vf_info *vfinfo = NULL;
	int rar_entry;
	struct rte_eth_dev *dev = NULL;
	struct rte_pci_device *pci_dev = NULL;

	RTE_ETH_VALID_PORTID_OR_ERR_RET(port, -ENODEV);

	dev = &rte_eth_devices[port];
	pci_dev = RTE_ETH_DEV_TO_PCI(dev);

	if (!is_txgbe_supported(dev))
		return -ENOTSUP;

	if (vf >= pci_dev->max_vfs)
		return -EINVAL;

	hw = TXGBE_DEV_HW(dev);
	vfinfo = *(TXGBE_DEV_VFDATA(dev));
	rar_entry = hw->mac.num_rar_entries - (vf + 1);

	if (is_valid_assigned_ether_addr(mac_addr)) {
		rte_memcpy(&vfinfo[vf].vf_mac_addr, mac_addr,
			   ETHER_ADDR_LEN);
		return hw->mac.set_rar(hw, rar_entry, (void *)mac_addr, vf,
					   true);
	}
	return -EINVAL;
}

int
rte_pmd_txgbe_ping_vf(uint16_t port, uint16_t vf)
{
	struct txgbe_hw *hw = NULL;
	struct txgbe_vf_info *vfinfo = NULL;
	struct rte_eth_dev *dev = NULL;
	struct rte_pci_device *pci_dev = NULL;
	uint32_t ctrl;

	RTE_ETH_VALID_PORTID_OR_ERR_RET(port, -ENODEV);

	dev = &rte_eth_devices[port];
	pci_dev = RTE_ETH_DEV_TO_PCI(dev);

	if (!is_txgbe_supported(dev))
		return -ENOTSUP;

	if (vf >= pci_dev->max_vfs)
		return -EINVAL;

	hw = TXGBE_DEV_HW(dev);
	vfinfo = *(TXGBE_DEV_VFDATA(dev));

	ctrl = TXGBE_PF_CONTROL_MSG;
	if (vfinfo[vf].clear_to_send)
		ctrl |= TXGBE_VT_MSGTYPE_CTS;

	txgbe_write_mbx(hw, &ctrl, 1, vf);

	return 0;
}

int
rte_pmd_txgbe_set_vf_vlan_anti_spoof(uint16_t port, uint16_t vf, uint8_t on)
{
	struct txgbe_hw *hw = NULL;
	struct txgbe_mac_info *mac = NULL;
	struct rte_eth_dev *dev = NULL;
	struct rte_pci_device *pci_dev = NULL;

	RTE_ETH_VALID_PORTID_OR_ERR_RET(port, -ENODEV);

	dev = &rte_eth_devices[port];
	pci_dev = RTE_ETH_DEV_TO_PCI(dev);

	if (!is_txgbe_supported(dev))
		return -ENOTSUP;

	if (vf >= pci_dev->max_vfs)
		return -EINVAL;

	if (on > 1)
		return -EINVAL;

	hw = TXGBE_DEV_HW(dev);
	mac = &hw->mac;

	mac->set_vlan_anti_spoofing(hw, on, vf);

	return 0;
}

int
rte_pmd_txgbe_set_vf_mac_anti_spoof(uint16_t port, uint16_t vf, uint8_t on)
{
	struct txgbe_hw *hw = NULL;
	struct txgbe_mac_info *mac = NULL;
	struct rte_eth_dev *dev = NULL;
	struct rte_pci_device *pci_dev = NULL;

	RTE_ETH_VALID_PORTID_OR_ERR_RET(port, -ENODEV);

	dev = &rte_eth_devices[port];
	pci_dev = RTE_ETH_DEV_TO_PCI(dev);

	if (!is_txgbe_supported(dev))
		return -ENOTSUP;

	if (vf >= pci_dev->max_vfs)
		return -EINVAL;

	if (on > 1)
		return -EINVAL;

	hw = TXGBE_DEV_HW(dev);
	mac = &hw->mac;
	mac->set_mac_anti_spoofing(hw, on, vf);

	return 0;
}

int
rte_pmd_txgbe_set_vf_vlan_insert(uint16_t port, uint16_t vf, uint16_t vlan_id)
{
	struct txgbe_hw *hw = NULL;
	uint32_t ctrl;
	struct rte_eth_dev *dev = NULL;
	struct rte_pci_device *pci_dev = NULL;

	RTE_ETH_VALID_PORTID_OR_ERR_RET(port, -ENODEV);

	dev = &rte_eth_devices[port];
	pci_dev = RTE_ETH_DEV_TO_PCI(dev);

	if (!is_txgbe_supported(dev))
		return -ENOTSUP;

	if (vf >= pci_dev->max_vfs)
		return -EINVAL;

	if (vlan_id > ETHER_MAX_VLAN_ID)
		return -EINVAL;

	hw = TXGBE_DEV_HW(dev);
	ctrl = rd32(hw, TXGBE_POOLTAG(vf));
	if (vlan_id) {
		ctrl = TXGBE_POOLTAG(vf);
		ctrl &= ~TXGBE_POOLTAG_ACT_MASK;
		ctrl |= TXGBE_POOLTAG_ACT_ALWAYS;

		ctrl &= ~TXGBE_POOLTAG_VTAG_MASK;
		ctrl |= TXGBE_POOLTAG_VTAG(vlan_id);
	} else {
		ctrl = 0;
	}

	wr32(hw, TXGBE_POOLTAG(vf), ctrl);

	return 0;
}

int
rte_pmd_txgbe_set_tx_loopback(uint16_t port, uint8_t on)
{
	struct txgbe_hw *hw = NULL;
	uint32_t ctrl;
	struct rte_eth_dev *dev = NULL;

	RTE_ETH_VALID_PORTID_OR_ERR_RET(port, -ENODEV);

	dev = &rte_eth_devices[port];

	if (!is_txgbe_supported(dev))
		return -ENOTSUP;

	if (on > 1)
		return -EINVAL;

	hw = TXGBE_DEV_HW(dev);
	ctrl = rd32(hw, TXGBE_PSRCTL);
	/* enable or disable VMDQ loopback */
	if (on)
		ctrl |= TXGBE_PSRCTL_LBENA;
	else
		ctrl &= ~TXGBE_PSRCTL_LBENA;

	wr32(hw, TXGBE_PSRCTL, ctrl);

	return 0;
}

int
rte_pmd_txgbe_set_all_queues_drop_en(uint16_t port, uint8_t on)
{
	struct txgbe_hw *hw = NULL;
	u16 i;
	int num_queues = 128;
	struct rte_eth_dev *dev = NULL;

	RTE_ETH_VALID_PORTID_OR_ERR_RET(port, -ENODEV);

	dev = &rte_eth_devices[port];

	if (!is_txgbe_supported(dev))
		return -ENOTSUP;

	if (on > 1)
		return -EINVAL;

	hw = TXGBE_DEV_HW(dev);
	for (i = 0; i <= num_queues; i++) {
		u32 val = 1 << (i % 32);
		wr32m(hw, TXGBE_QPRXDROP(i / 32), val, val);
	}

	return 0;
}

int
rte_pmd_txgbe_set_vf_split_drop_en(uint16_t port, uint16_t vf, uint8_t on)
{
	struct txgbe_hw *hw = NULL;
	uint32_t reg_value;
	struct rte_eth_dev *dev = NULL;
	struct rte_pci_device *pci_dev = NULL;

	RTE_ETH_VALID_PORTID_OR_ERR_RET(port, -ENODEV);

	dev = &rte_eth_devices[port];
	pci_dev = RTE_ETH_DEV_TO_PCI(dev);

	if (!is_txgbe_supported(dev))
		return -ENOTSUP;

	/* only support VF's 0 to 63 */
	if ((vf >= pci_dev->max_vfs) || (vf > 63))
		return -EINVAL;

	if (on > 1)
		return -EINVAL;

	hw = TXGBE_DEV_HW(dev);
	reg_value = rd32(hw, TXGBE_RXCFG(vf * 2));
	if (on)
		reg_value |= TXGBE_RXCFG_DROP;
	else
		reg_value &= ~TXGBE_RXCFG_DROP;

	wr32(hw, TXGBE_RXCFG(vf * 2), reg_value);

	reg_value = rd32(hw, TXGBE_RXCFG(vf * 2 + 1));
	if (on)
		reg_value |= TXGBE_RXCFG_DROP;
	else
		reg_value &= ~TXGBE_RXCFG_DROP;

	wr32(hw, TXGBE_RXCFG(vf * 2 + 1), reg_value);

	return 0;
}

int
rte_pmd_txgbe_set_vf_vlan_stripq(uint16_t port, uint16_t vf, uint8_t on)
{
	struct rte_eth_dev *dev = NULL;
	struct rte_pci_device *pci_dev = NULL;
	struct txgbe_hw *hw = NULL;
	uint16_t queues_per_pool;
	uint32_t q;

	RTE_ETH_VALID_PORTID_OR_ERR_RET(port, -ENODEV);

	dev = &rte_eth_devices[port];
	pci_dev = RTE_ETH_DEV_TO_PCI(dev);
	hw = TXGBE_DEV_HW(dev);

	if (!is_txgbe_supported(dev))
		return -ENOTSUP;

	if (vf >= pci_dev->max_vfs)
		return -EINVAL;

	if (on > 1)
		return -EINVAL;

	RTE_FUNC_PTR_OR_ERR_RET(*dev->dev_ops->vlan_strip_queue_set, -ENOTSUP);

	/* The PF has 128 queue pairs and in SRIOV configuration
	 * those queues will be assigned to VF's, so RXDCTL
	 * registers will be dealing with queues which will be
	 * assigned to VF's.
	 * Let's say we have SRIOV configured with 31 VF's then the
	 * first 124 queues 0-123 will be allocated to VF's and only
	 * the last 4 queues 123-127 will be assigned to the PF.
	 */
	queues_per_pool = (uint16_t)hw->mac.max_rx_queues /
			  ETH_64_POOLS;

	for (q = 0; q < queues_per_pool; q++)
		(*dev->dev_ops->vlan_strip_queue_set)(dev,
				q + vf * queues_per_pool, on);
	return 0;
}

int
rte_pmd_txgbe_set_vf_rxmode(uint16_t port, uint16_t vf,
			    uint16_t rx_mask, uint8_t on)
{
	u32 val = 0;
	struct rte_eth_dev *dev = NULL;
	struct rte_pci_device *pci_dev = NULL;
	struct txgbe_hw *hw = NULL;
	uint32_t vmolr;

	RTE_ETH_VALID_PORTID_OR_ERR_RET(port, -ENODEV);

	dev = &rte_eth_devices[port];
	pci_dev = RTE_ETH_DEV_TO_PCI(dev);

	if (!is_txgbe_supported(dev))
		return -ENOTSUP;

	if (vf >= pci_dev->max_vfs)
		return -EINVAL;

	if (on > 1)
		return -EINVAL;

	hw = TXGBE_DEV_HW(dev);
	vmolr = rd32(hw, TXGBE_POOLETHCTL(vf));

	if (txgbe_vt_check(hw) < 0)
		return -ENOTSUP;

	val = txgbe_convert_vm_rx_mask_to_val(rx_mask, val);

	if (on)
		vmolr |= val;
	else
		vmolr &= ~val;

	wr32(hw, TXGBE_POOLETHCTL(vf), vmolr);

	return 0;
}

int
rte_pmd_txgbe_set_vf_rx(uint16_t port, uint16_t vf, uint8_t on)
{
	struct rte_eth_dev *dev = NULL;
	struct rte_pci_device *pci_dev = NULL;
	uint32_t reg, addr;
	uint32_t val;
	const uint8_t bit1 = 0x1;
	struct txgbe_hw *hw = NULL;

	RTE_ETH_VALID_PORTID_OR_ERR_RET(port, -ENODEV);

	dev = &rte_eth_devices[port];
	pci_dev = RTE_ETH_DEV_TO_PCI(dev);

	if (!is_txgbe_supported(dev))
		return -ENOTSUP;

	if (vf >= pci_dev->max_vfs)
		return -EINVAL;

	if (on > 1)
		return -EINVAL;

	hw = TXGBE_DEV_HW(dev);

	if (txgbe_vt_check(hw) < 0)
		return -ENOTSUP;

	/* for vf >= 32, set bit in PFVFRE[1], otherwise PFVFRE[0] */
	if (vf >= 32) {
		addr = TXGBE_POOLRXENA(1);
		val = bit1 << (vf - 32);
	} else {
		addr = TXGBE_POOLRXENA(0);
		val = bit1 << vf;
	}

	reg = rd32(hw, addr);

	if (on)
		reg |= val;
	else
		reg &= ~val;

	wr32(hw, addr, reg);

	return 0;
}

int
rte_pmd_txgbe_set_vf_tx(uint16_t port, uint16_t vf, uint8_t on)
{
	struct rte_eth_dev *dev = NULL;
	struct rte_pci_device *pci_dev = NULL;
	uint32_t reg, addr;
	uint32_t val;
	const uint8_t bit1 = 0x1;

	struct txgbe_hw *hw = NULL;

	RTE_ETH_VALID_PORTID_OR_ERR_RET(port, -ENODEV);

	dev = &rte_eth_devices[port];
	pci_dev = RTE_ETH_DEV_TO_PCI(dev);

	if (!is_txgbe_supported(dev))
		return -ENOTSUP;

	if (vf >= pci_dev->max_vfs)
		return -EINVAL;

	if (on > 1)
		return -EINVAL;

	hw = TXGBE_DEV_HW(dev);
	if (txgbe_vt_check(hw) < 0)
		return -ENOTSUP;

	/* for vf >= 32, set bit in PFVFTE[1], otherwise PFVFTE[0] */
	if (vf >= 32) {
		addr = TXGBE_POOLTXENA(1);
		val = bit1 << (vf - 32);
	} else {
		addr = TXGBE_POOLTXENA(0);
		val = bit1 << vf;
	}

	reg = rd32(hw, addr);

	if (on)
		reg |= val;
	else
		reg &= ~val;

	wr32(hw, addr, reg);

	return 0;
}

int
rte_pmd_txgbe_set_vf_vlan_filter(uint16_t port, uint16_t vlan,
				 uint64_t vf_mask, uint8_t vlan_on)
{
	struct rte_eth_dev *dev = NULL;
	int ret = 0;
	uint16_t vf_idx;
	struct txgbe_hw *hw = NULL;

	RTE_ETH_VALID_PORTID_OR_ERR_RET(port, -ENODEV);

	dev = &rte_eth_devices[port];

	if (!is_txgbe_supported(dev))
		return -ENOTSUP;

	if ((vlan > ETHER_MAX_VLAN_ID) || (vf_mask == 0))
		return -EINVAL;

	hw = TXGBE_DEV_HW(dev);
	if (txgbe_vt_check(hw) < 0)
		return -ENOTSUP;

	for (vf_idx = 0; vf_idx < 64; vf_idx++) {
		if (vf_mask & ((uint64_t)(1ULL << vf_idx))) {
			ret = hw->mac.set_vfta(hw, vlan, vf_idx,
						   vlan_on, false);
			if (ret < 0)
				return ret;
		}
	}

	return ret;
}

int
rte_pmd_txgbe_set_vf_rate_limit(uint16_t port, uint16_t vf,
				uint16_t tx_rate, uint64_t q_msk)
{
	struct rte_eth_dev *dev = NULL;

	RTE_ETH_VALID_PORTID_OR_ERR_RET(port, -ENODEV);

	dev = &rte_eth_devices[port];

	if (!is_txgbe_supported(dev))
		return -ENOTSUP;

	return txgbe_set_vf_rate_limit(dev, vf, tx_rate, q_msk);
}

int
rte_pmd_txgbe_macsec_enable(uint16_t port, uint8_t en, uint8_t rp)
{
	struct rte_eth_dev *dev = NULL;
	struct txgbe_macsec_setting macsec_setting;

	RTE_ETH_VALID_PORTID_OR_ERR_RET(port, -ENODEV);

	dev = &rte_eth_devices[port];

	macsec_setting.offload_en = 1;
	macsec_setting.encrypt_en = en;
	macsec_setting.replayprotect_en = rp;

	txgbe_dev_macsec_setting_save(dev, &macsec_setting);

	txgbe_dev_macsec_register_enable(dev, &macsec_setting);

	return 0;
}

int
rte_pmd_txgbe_macsec_disable(uint16_t port)
{
	struct rte_eth_dev *dev = NULL;

	RTE_ETH_VALID_PORTID_OR_ERR_RET(port, -ENODEV);

	dev = &rte_eth_devices[port];

	txgbe_dev_macsec_setting_reset(dev);

	txgbe_dev_macsec_register_disable(dev);

	return 0;
}

int
rte_pmd_txgbe_macsec_config_txsc(uint16_t port, uint8_t *mac)
{
	struct txgbe_hw *hw = NULL;
	struct rte_eth_dev *dev = NULL;
	uint32_t ctrl;

	RTE_ETH_VALID_PORTID_OR_ERR_RET(port, -ENODEV);

	dev = &rte_eth_devices[port];

	if (!is_txgbe_supported(dev))
		return -ENOTSUP;

	hw = TXGBE_DEV_HW(dev);

	ctrl = mac[0] | (mac[1] << 8) | (mac[2] << 16) | (mac[3] << 24);
	wr32(hw, TXGBE_LSECTXSCIL, ctrl);

	ctrl = mac[4] | (mac[5] << 8);
	wr32(hw, TXGBE_LSECTXSCIH, ctrl);

	return 0;
}

int
rte_pmd_txgbe_macsec_config_rxsc(uint16_t port, uint8_t *mac, uint16_t pi)
{
	struct txgbe_hw *hw = NULL;
	struct rte_eth_dev *dev = NULL;
	uint32_t ctrl;

	RTE_ETH_VALID_PORTID_OR_ERR_RET(port, -ENODEV);

	dev = &rte_eth_devices[port];

	if (!is_txgbe_supported(dev))
		return -ENOTSUP;

	hw = TXGBE_DEV_HW(dev);

	ctrl = mac[0] | (mac[1] << 8) | (mac[2] << 16) | (mac[3] << 24);
	wr32(hw, TXGBE_LSECRXSCIL, ctrl);

	pi = rte_cpu_to_be_16(pi);
	ctrl = mac[4] | (mac[5] << 8) | (pi << 16);
	wr32(hw, TXGBE_LSECRXSCIH, ctrl);

	return 0;
}

int
rte_pmd_txgbe_macsec_select_txsa(uint16_t port, uint8_t idx, uint8_t an,
				 uint32_t pn, uint8_t *key)
{
	struct txgbe_hw *hw = NULL;
	struct rte_eth_dev *dev = NULL;
	uint32_t ctrl, i;

	RTE_ETH_VALID_PORTID_OR_ERR_RET(port, -ENODEV);

	dev = &rte_eth_devices[port];

	if (!is_txgbe_supported(dev))
		return -ENOTSUP;

	hw = TXGBE_DEV_HW(dev);

	if (idx != 0 && idx != 1)
		return -EINVAL;

	if (an >= 4)
		return -EINVAL;

	hw = TXGBE_DEV_HW(dev);

	/* Set the PN and key */
	pn = rte_cpu_to_be_32(pn);
	if (idx == 0) {
		wr32(hw, TXGBE_LSECTXPN0, pn);

		for (i = 0; i < 4; i++) {
			ctrl = (key[i * 4 + 0] <<  0) |
			       (key[i * 4 + 1] <<  8) |
			       (key[i * 4 + 2] << 16) |
			       (key[i * 4 + 3] << 24);
			wr32(hw, TXGBE_LSECTXKEY0(i), ctrl);
		}
	} else {
		wr32(hw, TXGBE_LSECTXPN1, pn);

		for (i = 0; i < 4; i++) {
			ctrl = (key[i * 4 + 0] <<  0) |
			       (key[i * 4 + 1] <<  8) |
			       (key[i * 4 + 2] << 16) |
			       (key[i * 4 + 3] << 24);
			wr32(hw, TXGBE_LSECTXKEY1(i), ctrl);
		}
	}

	/* Set AN and select the SA */
	ctrl = (an << idx * 2) | (idx << 4);
	wr32(hw, TXGBE_LSECTXSA, ctrl);

	return 0;
}

int
rte_pmd_txgbe_macsec_select_rxsa(uint16_t port, uint8_t idx, uint8_t an,
				 uint32_t pn, uint8_t *key)
{
	struct txgbe_hw *hw = NULL;
	struct rte_eth_dev *dev = NULL;
	uint32_t ctrl, i;

	RTE_ETH_VALID_PORTID_OR_ERR_RET(port, -ENODEV);

	dev = &rte_eth_devices[port];

	if (!is_txgbe_supported(dev))
		return -ENOTSUP;

	hw = TXGBE_DEV_HW(dev);

	if (idx != 0 && idx != 1)
		return -EINVAL;

	if (an >= 4)
		return -EINVAL;

	/* Set the PN */
	pn = rte_cpu_to_be_32(pn);
	wr32(hw, TXGBE_LSECRXPN(idx), pn);

	/* Set the key */
	for (i = 0; i < 4; i++) {
		ctrl = (key[i * 4 + 0] <<  0) |
		       (key[i * 4 + 1] <<  8) |
		       (key[i * 4 + 2] << 16) |
		       (key[i * 4 + 3] << 24);
		wr32(hw, TXGBE_LSECRXKEY(idx, i), ctrl);
	}

	/* Set the AN and validate the SA */
	ctrl = an | (1 << 2);
	wr32(hw, TXGBE_LSECRXSA(idx), ctrl);

	return 0;
}

int
rte_pmd_txgbe_set_tc_bw_alloc(uint16_t port,
			      uint8_t tc_num,
			      uint8_t *bw_weight)
{
	struct rte_eth_dev *dev = NULL;
	struct txgbe_dcb_config *dcb_config = NULL;
	struct txgbe_dcb_tc_config *tc = NULL;
	struct rte_eth_conf *eth_conf = NULL;
	struct txgbe_bw_conf *bw_conf = NULL;
	uint8_t i;
	uint8_t nb_tcs;
	uint16_t sum;

	RTE_ETH_VALID_PORTID_OR_ERR_RET(port, -ENODEV);

	dev = &rte_eth_devices[port];

	if (!is_txgbe_supported(dev))
		return -ENOTSUP;

	if (tc_num > TXGBE_DCB_TC_MAX) {
		PMD_DRV_LOG(ERR, "TCs should be no more than %d.",
			    TXGBE_DCB_TC_MAX);
		return -EINVAL;
	}

	dcb_config = TXGBE_DEV_DCB_CONFIG(dev);
	bw_conf = TXGBE_DEV_BW_CONF(dev);
	eth_conf = &dev->data->dev_conf;

	if (eth_conf->txmode.mq_mode == ETH_MQ_TX_DCB) {
		nb_tcs = eth_conf->tx_adv_conf.dcb_tx_conf.nb_tcs;
	} else if (eth_conf->txmode.mq_mode == ETH_MQ_TX_VMDQ_DCB) {
		if (eth_conf->tx_adv_conf.vmdq_dcb_tx_conf.nb_queue_pools ==
		    ETH_32_POOLS)
			nb_tcs = ETH_4_TCS;
		else
			nb_tcs = ETH_8_TCS;
	} else {
		nb_tcs = 1;
	}

	if (nb_tcs != tc_num) {
		PMD_DRV_LOG(ERR,
			    "Weight should be set for all %d enabled TCs.",
			    nb_tcs);
		return -EINVAL;
	}

	sum = 0;
	for (i = 0; i < nb_tcs; i++)
		sum += bw_weight[i];
	if (sum != 100) {
		PMD_DRV_LOG(ERR,
			    "The summary of the TC weight should be 100.");
		return -EINVAL;
	}

	for (i = 0; i < nb_tcs; i++) {
		tc = &dcb_config->tc_config[i];
		tc->path[TXGBE_DCB_TX_CONFIG].bwg_percent = bw_weight[i];
	}
	for (; i < TXGBE_DCB_TC_MAX; i++) {
		tc = &dcb_config->tc_config[i];
		tc->path[TXGBE_DCB_TX_CONFIG].bwg_percent = 0;
	}

	bw_conf->tc_num = nb_tcs;

	return 0;
}

int __rte_experimental
rte_pmd_txgbe_upd_fctrl_sbp(uint16_t port, int enable)
{
	struct txgbe_hw *hw = NULL;
	struct rte_eth_dev *dev = NULL;
	uint32_t fctrl;

	RTE_ETH_VALID_PORTID_OR_ERR_RET(port, -ENODEV);
	dev = &rte_eth_devices[port];
	if (!is_txgbe_supported(dev))
		return -ENOTSUP;

	hw = TXGBE_DEV_HW(dev);
	if (!hw)
		return -ENOTSUP;

	fctrl = rd32(hw, TXGBE_SECRXCTL);

	/* If 'enable' set the SBP bit else clear it */
	if (enable) {
		fctrl |= TXGBE_SECRXCTL_SAVEBAD;
		hw->mode = 1;
		wr32m(hw, TXGBE_FRMSZ, TXGBE_FRMSZ_MAX_MASK, TXGBE_FRAME_SIZE_MAX);
	} else {
		fctrl &= ~(TXGBE_SECRXCTL_SAVEBAD);
		hw->mode = 0;
		wr32m(hw, TXGBE_FRMSZ, TXGBE_FRMSZ_MAX_MASK,
			TXGBE_FRMSZ_MAX(dev->data->dev_conf.rxmode.max_rx_pkt_len));
	}
	wr32(hw, TXGBE_SECRXCTL, fctrl);
	return 0;
}

/**
 *  rte_pmd_txgbe_acquire_swfw - Acquire SWFW semaphore
 *  @hw: pointer to hardware structure
 *  @mask: Mask to specify which semaphore to acquire
 *
 *  Acquires the SWFW semaphore and get the shared phy token as needed
 */
STATIC s32 rte_pmd_txgbe_acquire_swfw(struct txgbe_hw *hw, u32 mask)
{
	int retries = FW_PHY_TOKEN_RETRIES;
	s32 status = 0;

	while (--retries) {
		status = hw->mac.acquire_swfw_sync(hw, mask);
		if (status) {
			PMD_DRV_LOG(ERR, "Get SWFW sem failed, Status = %d\n",
				    status);
			return status;
		}
		if (status == 0)
			return 0;

		if (status == TXGBE_ERR_TOKEN_RETRY)
			PMD_DRV_LOG(ERR, "Get PHY token failed, Status = %d\n",
				    status);

		hw->mac.release_swfw_sync(hw, mask);
		if (status != TXGBE_ERR_TOKEN_RETRY) {
			PMD_DRV_LOG(ERR,
				    "Retry get PHY token failed, Status=%d\n",
				    status);
			return status;
		}
	}
	PMD_DRV_LOG(ERR, "swfw acquisition retries failed!: PHY ID = 0x%08X\n",
		    hw->phy.id);
	return status;
}

/**
 *  rte_pmd_txgbe_release_swfw_sync - Release SWFW semaphore
 *  @hw: pointer to hardware structure
 *  @mask: Mask to specify which semaphore to release
 *
 *  Releases the SWFW semaphore and puts the shared phy token as needed
 */
STATIC void rte_pmd_txgbe_release_swfw(struct txgbe_hw *hw, u32 mask)
{
	hw->mac.release_swfw_sync(hw, mask);
}

int __rte_experimental
rte_pmd_txgbe_mdio_lock(uint16_t port)
{
	struct txgbe_hw *hw = NULL;
	struct rte_eth_dev *dev = NULL;
	u32 swfw_mask;

	RTE_ETH_VALID_PORTID_OR_ERR_RET(port, -ENODEV);
	dev = &rte_eth_devices[port];
	if (!is_txgbe_supported(dev))
		return -ENOTSUP;

	hw = TXGBE_DEV_HW(dev);
	if (!hw)
		return -ENOTSUP;

	if (hw->bus.lan_id)
		swfw_mask = TXGBE_MNGSEM_SWPHY;
	else
		swfw_mask = TXGBE_MNGSEM_SWPHY;

	if (rte_pmd_txgbe_acquire_swfw(hw, swfw_mask))
		return TXGBE_ERR_SWFW_SYNC;

	return 0;
}

int __rte_experimental
rte_pmd_txgbe_mdio_unlock(uint16_t port)
{
	struct rte_eth_dev *dev = NULL;
	struct txgbe_hw *hw = NULL;
	u32 swfw_mask;

	RTE_ETH_VALID_PORTID_OR_ERR_RET(port, -ENODEV);

	dev = &rte_eth_devices[port];
	if (!is_txgbe_supported(dev))
		return -ENOTSUP;

	hw = TXGBE_DEV_HW(dev);
	if (!hw)
		return -ENOTSUP;

	if (hw->bus.lan_id)
		swfw_mask = TXGBE_MNGSEM_SWPHY;
	else
		swfw_mask = TXGBE_MNGSEM_SWPHY;

	rte_pmd_txgbe_release_swfw(hw, swfw_mask);

	return 0;
}

int __rte_experimental
rte_pmd_txgbe_mdio_unlocked_read(uint16_t port, uint32_t reg_addr,
				 uint32_t dev_type, uint16_t *phy_data)
{
	struct txgbe_hw *hw = NULL;
	struct rte_eth_dev *dev = NULL;
	u32 command;

	RTE_ETH_VALID_PORTID_OR_ERR_RET(port, -ENODEV);
	dev = &rte_eth_devices[port];
	if (!is_txgbe_supported(dev))
		return -ENOTSUP;

	hw = TXGBE_DEV_HW(dev);
	if (!hw)
		return -ENOTSUP;

	/* write command */
	command = TXGBE_MDIOSCA_REG(reg_addr) |
		  TXGBE_MDIOSCA_DEV(dev_type) |
		  TXGBE_MDIOSCA_PORT(hw->phy.addr);
	wr32(hw, TXGBE_MDIOSCA, command);

	command = TXGBE_MDIOSCD_CMD_READ |
		  TXGBE_MDIOSCD_BUSY;
	wr32(hw, TXGBE_MDIOSCD, command);

	/* wait for completion */
	if (!po32m(hw, TXGBE_MDIOSCD, TXGBE_MDIOSCD_BUSY,
		0, NULL, 100, 100)) {
		TLOG_DEBUG("PHY read command did not complete.\n");
		return -TERR_PHY;
	}

	*phy_data = rd32m(hw, TXGBE_MDIOSCA, TXGBE_MDIOSCD_DAT(~0));

	return 0;
}

int __rte_experimental
rte_pmd_txgbe_mdio_unlocked_write(uint16_t port, uint32_t reg_addr,
				  uint32_t dev_type, uint16_t phy_data)
{
	struct txgbe_hw *hw = NULL;
	u32 command;
	struct rte_eth_dev *dev = NULL;

	RTE_ETH_VALID_PORTID_OR_ERR_RET(port, -ENODEV);
	dev = &rte_eth_devices[port];
	if (!is_txgbe_supported(dev))
		return -ENOTSUP;

	hw = TXGBE_DEV_HW(dev);
	if (!hw)
		return -ENOTSUP;

	/* Put the data in the MDI single read and write data register*/
	command = TXGBE_MDIOSCA_REG(reg_addr) |
		  TXGBE_MDIOSCA_DEV(dev_type) |
		  TXGBE_MDIOSCA_PORT(hw->phy.addr);
	wr32(hw, TXGBE_MDIOSCA, command);

	command = TXGBE_MDIOSCD_CMD_WRITE |
		  TXGBE_MDIOSCD_DAT(phy_data) |
		  TXGBE_MDIOSCD_BUSY;
	wr32(hw, TXGBE_MDIOSCD, command);

	/* wait for completion */
	if (!po32m(hw, TXGBE_MDIOSCD, TXGBE_MDIOSCD_BUSY,
		0, NULL, 100, 100)) {
		TLOG_DEBUG("PHY write cmd didn't complete\n");
		return -TERR_PHY;
	}

	return 0;
}
